/* matriz.c - implementacao das rotinas para matrizes esparcas */

#include <stdio.h>
#include <stdlib.h>
#include <process.h>

#include "matriz.h"
#include "alocacao.h"

/*---------------------------------------------------------------------------*\
 | InsereNodo - Insere qualquer nodo na matriz fornecida, inclusive os nodos |
 |              cabecas das linha e das colunas.                             |
 |									     |
 | Recebe - Um apontador para o nodo principal da matriz ja alocado, um outro|
 |	    apontador para o nodo a ser incluido e a sua posicao na linha e  |
 |	    na coluna.							     |
 |                                                                           |
 | Retorna - Nada					                     |
\*---------------------------------------------------------------------------*/

void InsereNodo (Apontador Matriz, Apontador Nodo, int m, int n)
{
  Apontador Linha, Coluna;
  int cont;

  if (m == -1) { /* Inserir um nodo na linha base (caso especial) */
    if (Matriz->Direita == NULL) { /* Primeiro nodo da fila da linha */
      Matriz->Direita = Nodo;
      Nodo->Direita = Matriz;
      Nodo->Abaixo = Nodo;
      Nodo->Coluna = -1;
    }
    else { /* Existe mais nodos na linha base */
      Nodo->Direita = Matriz->Direita;
      Matriz->Direita = Nodo;
      Nodo->Abaixo = Nodo;
      Nodo->Coluna = -1;
    }
    return;
  }

  if (n == -1) { /* Inserir um nodo na coluna base */
    if (Matriz->Abaixo == NULL) { /* Primeiro nodo da fila da coluna */
      Matriz->Abaixo = Nodo;
      Nodo->Abaixo = Matriz;
      Nodo->Direita = Nodo;
      Nodo->Linha = -1;
    }
    else { /* Existe mais nodos na coluna base */
      Nodo->Abaixo = Matriz->Abaixo;
      Matriz->Abaixo = Nodo;
      Nodo->Direita = Nodo;
      Nodo->Linha = -1;
    }
    return;
  }

  Nodo->Linha = m; Nodo->Coluna = n;

  Linha = Coluna = Matriz;

  for (cont = 0; cont < m; cont++) /* Avanca para a linha certa */
    Linha = Linha->Abaixo;

  for (cont = 0; cont < n; cont++) /* Avanca para a coluna certa */
    Coluna = Coluna->Direita;

  while (Linha->Direita->Linha != -1 &&         /* Se desloca para o nodo */
	 Linha->Direita->Coluna < Nodo->Coluna) /* antes da posicao certa */
    Linha = Linha->Direita;                     /* na linha */

  while (Coluna->Abaixo->Coluna != -1 &&        /* Se desloca para o nodo */
	 Coluna->Abaixo->Linha < Nodo->Linha)   /* antes da posicao certa */
    Coluna = Coluna->Abaixo;                    /* na coluna */

  Nodo->Direita = Linha->Direita;               /* Insere na linha */
  Linha->Direita = Nodo;

  Nodo->Abaixo = Coluna->Abaixo;                /* Insere na coluna */
  Coluna->Abaixo = Nodo;

}

/*---------------------------------------------------------------------------*\
 | AlocaEstruturaBasicaMatriz - Monta a estrutura dos nodos cabeca da matriz.|
 |									     |
 | Recebe - O numero de linha e de colunas da matriz.                        |
 |	                                                                     |
 | Retorna - O endereco da matriz ja construida                              |
\*---------------------------------------------------------------------------*/

Apontador AlocaEstruturaBasicaMatriz (int m, int n)
{
  Apontador NovaMatriz, Nodo;

  int TotalNodos, Cont;

  NovaMatriz = AlocaNodo ();                   /* Nodo cabeca principal    */
  NovaMatriz->Linha = NovaMatriz->Coluna = -1; /* Valor Linha = -1 e       */
					       /* Coluna = -1 e usado na   */
					       /* Na funcao ApagaMatriz    */
					       /* p/ localizacao.          */

  for (Cont = 0; Cont < n; Cont++) {           /* Constroi os nodos cabeca */
    Nodo = AlocaNodo ();                       /* das linhas               */
    InsereNodo (NovaMatriz, Nodo, -1, Cont);
  }

  for (Cont = 0; Cont < m; Cont++) {           /* Constroi os nodos cabeca */
    Nodo = AlocaNodo ();                       /* das colunas              */
    InsereNodo (NovaMatriz, Nodo, Cont, -1);
  }

  return NovaMatriz;
}

/*---------------------------------------------------------------------------*\
 | LeMatriz - Le o arquivo da matriz e manda contruir a estrutura na memoria,|
 |            retornando entao o ponteiro para a matriz ja alocada.          |
 |									     |
 | recebe - O nome do arquivo no DOS.                                        |
 |	                                                                     |
 | retorna - A estrutura da matriz na memoria.                               |
\*---------------------------------------------------------------------------*/

Apontador LeMatriz (char *NomeArquivo)
{
  FILE *ArquivoLeitura;

  int m, n; /* Numero de linha e colunas da matriz */

  float valor;

  Apontador NovaMatriz, Nodo;

  if ((ArquivoLeitura = fopen (NomeArquivo, "r")) == NULL) {
    printf ("Erro: Arquivo %s\n", NomeArquivo);
    printf ("Nao foi possivel abri-lo.\n");
    exit (EXIT_FAILURE);
  }

  if (fscanf (ArquivoLeitura, "%d, %d\n", &m, &n) != 2) {
    printf ("Erro: Arquivo %s\n", NomeArquivo);
    printf ("O numero de linha e coluna da matriz nao foram encontrados\n");
    exit (EXIT_FAILURE);
  }

  NovaMatriz = AlocaEstruturaBasicaMatriz (m, n);

  while (!feof (ArquivoLeitura)) {
    if (fscanf (ArquivoLeitura, "%d, %d, %f\n", &m, &n, &valor) != 3) {
      printf ("Erro: Arquivo %s\n", NomeArquivo);
      printf ("Elemento da matriz nao encontrado.\n");
      exit (EXIT_FAILURE);
    }
    if (valor != 0l) {
      Nodo = AlocaNodo ();
      Nodo->Valor = valor;
      InsereNodo (NovaMatriz, Nodo, m, n);
    }
  }

  return NovaMatriz;
}

/*---------------------------------------------------------------------------*\
 | ApagaLinha - Funcao recursiva que desaloca uma linha da matriz.           |
 |									     |
 | recebe - O nodo cabeca da linha.                                          |
 |	                                                                     |
 | retorna - Nada.                                                           |
\*---------------------------------------------------------------------------*/

void ApagaLinha (Apontador Nodo)
{
  if (Nodo->Direita->Linha != -1)
    ApagaLinha (Nodo->Direita);
  EliminaNodo (Nodo);
}

/*---------------------------------------------------------------------------*\
 | ApagaLinha - Funcao recursiva que desaloca a matriz.                      |
 |									     |
 | recebe - Um ponteiro para a matriz.                                       |
 |	                                                                     |
 | retorna - Nada.                                                           |
\*---------------------------------------------------------------------------*/

void ApagaMatriz (Apontador Matriz)
{
  if (Matriz->Abaixo->Coluna != -1)
    ApagaMatriz (Matriz->Abaixo);
  ApagaLinha (Matriz);
}

/*---------------------------------------------------------------------------*\
 | LeNodo - Le o valor de uma posicao dentro da matriz.                      |
 |                                                                           |
 | Recebe - Um ponteiro para a matriz, a linha e a coluna desejada.          |
 |                                                                           |
 | Retorna - O valor lido.						     |
\*---------------------------------------------------------------------------*/

float LeNodo (Apontador Matriz, int m, int n)
{
  Apontador Nodo;

  int cont;

  Nodo = Matriz;

  for (cont = 0; cont < m; cont++) /* Se posiciona na linha correta */
    Nodo = Nodo->Abaixo;

  Nodo = Nodo->Direita;
  while (Nodo->Linha != -1 && Nodo->Coluna <= n) { /* Procura a coluna */
    if (Nodo->Linha == m && Nodo->Coluna == n)
      return Nodo->Valor;
    Nodo = Nodo->Direita;
  }

  return 0l;
}

/*---------------------------------------------------------------------------*\
 | ImprimeMatriz - Mostra a matriz na saida stdout.                          |
 |                                                                           |
 | Recebe - Uma mensagem identificando a matriz que sera impressa e um       |
 |          ponteiro para a matriz na memoria.                               |
 |                                                                           |
 | Retorna - Nada.       						     |
\*---------------------------------------------------------------------------*/

void ImprimeMatriz (char *Mensagem, Apontador Matriz)
{

  Apontador NodoLinha, NodoColuna;

  int m, n,
      linha, coluna;

  for (NodoLinha = Matriz->Abaixo, m = 0;       /* Conta quantas linha */
       NodoLinha->Coluna != -1;
       m++, NodoLinha = NodoLinha->Abaixo);

  for (NodoColuna = Matriz->Direita, n = 0;     /* Conta quantas colunas */
       NodoColuna->Linha != -1;
       n++, NodoColuna = NodoColuna->Direita);

  printf ("\n%s\n", Mensagem);

  for (linha = 1; linha <= m; linha++) {
    for (coluna = 1; coluna <= n; coluna++)
      printf ("%f ", LeNodo (Matriz, linha, coluna));
    putchar ('\n');
  }
}

/*---------------------------------------------------------------------------*\
 | SomaMatriz - Soma duas matrizes.                                          |
 |                                                                           |
 | recebe - dois apontadores para as duas matrizes que serao somadas.        |
 |                                                                           |
 | retorna - um apontador para a matriz soma.                                |
\*---------------------------------------------------------------------------*/

Apontador SomaMatriz (Apontador MatrizA, Apontador MatrizB)
{

  Apontador Nodo,
	    NodoLinha, NodoColuna,
	    MatrizC;

  float valor;

  int LinMatrizA, ColMatrizA,
      LinMatrizB, ColMatrizB,
      linha, coluna;

  for (NodoLinha = MatrizA->Abaixo, LinMatrizA = 0; /* Numero de linhas de A */
       NodoLinha->Coluna != -1;
       LinMatrizA++, NodoLinha = NodoLinha->Abaixo);

  for (NodoColuna = MatrizA->Direita, ColMatrizA = 0;/* Num. de colunas de A */
       NodoColuna->Linha != -1;
       ColMatrizA++, NodoColuna = NodoColuna->Direita);

  for (NodoLinha = MatrizB->Abaixo, LinMatrizB = 0; /* Numero de linhas de B */
       NodoLinha->Coluna != -1;
       LinMatrizB++, NodoLinha = NodoLinha->Abaixo);

  for (NodoColuna = MatrizB->Direita, ColMatrizB = 0;/* Mum. de colunas de A */
       NodoColuna->Linha != -1;
       ColMatrizB++, NodoColuna = NodoColuna->Direita);

  if (LinMatrizA != LinMatrizB || ColMatrizA != ColMatrizB) {
    printf ("Erro: Soma de matrizes\n");
    printf ("O tamanho das matrizes nao e igual.\n");
    exit (EXIT_FAILURE);
  }

  MatrizC = AlocaEstruturaBasicaMatriz (LinMatrizA, ColMatrizA);

  for (linha = 1; linha <= LinMatrizA; linha++)
    for (coluna = 1; coluna <= ColMatrizB; coluna++) {
      valor = LeNodo (MatrizA, linha, coluna) +
	      LeNodo (MatrizB, linha, coluna);
      if (valor != 0l) {
	Nodo = AlocaNodo ();
	Nodo->Valor = valor;
	InsereNodo (MatrizC, Nodo, linha, coluna);
      }
    }

  return MatrizC;

}

/*---------------------------------------------------------------------------*\
 | MultiplicaMatriz - Multiplica duas matrizes.				     |
 |                                                                           |
 | recebe - dois apontadores para duas matrizes que serao multiplicadas.     |
 |                                                                           |
 | retorna - um apontador para a matriz resultado.                           |
\*---------------------------------------------------------------------------*/

Apontador MultiplicaMatriz (Apontador MatrizA, Apontador MatrizB)
{

  Apontador Nodo,
	    NodoLinha, NodoColuna,
	    MatrizC;

  float valor;

  int LinMatrizA, ColMatrizA,
      LinMatrizB, ColMatrizB,
      linha, coluna, cont;

  for (NodoLinha = MatrizA->Abaixo, LinMatrizA = 0; /* Numero de linhas de A */
       NodoLinha->Coluna != -1;
       LinMatrizA++, NodoLinha = NodoLinha->Abaixo);

  for (NodoColuna = MatrizA->Direita, ColMatrizA = 0;/* Num. de colunas de A */
       NodoColuna->Linha != -1;
       ColMatrizA++, NodoColuna = NodoColuna->Direita);

  for (NodoLinha = MatrizB->Abaixo, LinMatrizB = 0; /* Numero de linhas de B */
       NodoLinha->Coluna != -1;
       LinMatrizB++, NodoLinha = NodoLinha->Abaixo);

  for (NodoColuna = MatrizB->Direita, ColMatrizB = 0;/* Num. de colunas de B */
       NodoColuna->Linha != -1;
       ColMatrizB++, NodoColuna = NodoColuna->Direita);

  if (ColMatrizA != LinMatrizB) {              /* Restricao de multiplicacao */
    printf ("Erro: Multiplicacao de matrizes\n");
    printf ("Numero de colunas nao bate com o numero de linhas\n");
    exit (EXIT_FAILURE);
  }

  MatrizC = AlocaEstruturaBasicaMatriz (LinMatrizA, ColMatrizB);

  for (linha = 1; linha <= LinMatrizA; linha++)
    for (coluna = 1; coluna <= ColMatrizB; coluna++) {
      for (cont = 1, valor = 0l; cont <= ColMatrizA; cont++)
	valor += LeNodo (MatrizA, linha, cont) *
		 LeNodo (MatrizB, cont, coluna);
      if (valor != 0l) {
	Nodo = AlocaNodo ();
	Nodo->Valor = valor;
	InsereNodo (MatrizC, Nodo, linha, coluna);
      }
    }

  return MatrizC;
}

/************\
 *   MAIN   *
\************/

void main (void)
{
  Apontador MatrizA, MatrizB, MatrizC;

  MatrizA = LeMatriz ("A.Mat");
  ImprimeMatriz ("Matriz A", MatrizA);
  MatrizB = LeMatriz ("B.Mat");
  ImprimeMatriz ("Matriz B", MatrizB);

  MatrizC = SomaMatriz (MatrizA, MatrizB);
  ImprimeMatriz ("Matriz C = Matriz A + Matriz B", MatrizC);
  ApagaMatriz (MatrizC);

  MatrizC = MultiplicaMatriz (MatrizA, MatrizB);
  ImprimeMatriz ("Matriz C = Matriz A X Matriz B", MatrizC);
  ApagaMatriz (MatrizB);
  ApagaMatriz (MatrizC);

  MatrizB = LeMatriz ("B.Mat");
  ImprimeMatriz ("Matriz A", MatrizA);
  ImprimeMatriz ("Matriz B", MatrizB);

  MatrizC = SomaMatriz (MatrizA, MatrizB);
  ImprimeMatriz ("Matriz C = Matriz A + Matriz B", MatrizC);
  ApagaMatriz (MatrizC);

  MatrizC = MultiplicaMatriz (MatrizB, MatrizA);
  ImprimeMatriz ("Matriz C = Matriz B * Matriz A", MatrizC);
  ApagaMatriz (MatrizC);

  MatrizC = MultiplicaMatriz (MatrizB, MatrizB);

  ImprimeMatriz ("Matriz A", MatrizA);
  ImprimeMatriz ("Matriz B", MatrizB);
  ImprimeMatriz ("Matriz C = Matriz B X Matriz B", MatrizC);

  ApagaMatriz (MatrizA);
  ApagaMatriz (MatrizB);
  ApagaMatriz (MatrizC);
}

